@@ -1,6 +1,7 @@
 /*
-   Copyright (C) 2002-2012, 2014-2017 Free Software Foundation, Inc.
-   Written by Keisuke Nishida, Roger While, Simon Sobisch, Ron Norman
+   Copyright (C) 2002-2012, 2014-2020 Free Software Foundation, Inc.
+   Written by Keisuke Nishida, Roger While, Simon Sobisch, Ron Norman,
+   Edwart Hard
 
    This file is part of GnuCOBOL.
 
@@ -15,11 +16,11 @@
    GNU Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General Public License
-   along with GnuCOBOL.  If not, see <http://www.gnu.org/licenses/>.
+   along with GnuCOBOL.  If not, see <https://www.gnu.org/licenses/>.
 */
 
 
-#include "config.h"
+#include <config.h>
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -35,7 +36,6 @@
 
 /* Force symbol exports */
 #define	COB_LIB_EXPIMP
-
 #include "libcob.h"
 #include "coblocal.h"
 
@@ -53,6 +53,16 @@ static const cob_field_attr	const_alpha_attr =
 static const cob_field_attr	const_binll_attr =
 				{COB_TYPE_NUMERIC_BINARY, 20, 0,
 				 COB_FLAG_HAVE_SIGN, NULL};
+static const cob_field_attr	all_display_attr =
+				{COB_TYPE_ALPHANUMERIC, 0, 0, 0, NULL};
+static const cob_field_attr	all_numeric_display_attr =
+				{COB_TYPE_NUMERIC_DISPLAY, COB_MAX_DIGITS, 0,
+				 0, NULL};
+
+static unsigned char all_numeric_data[COB_MAX_DIGITS];
+static const cob_field	all_numeric_field  = 
+				{COB_MAX_DIGITS, all_numeric_data,
+				 &all_numeric_display_attr};
 
 static const int	cob_exp10[10] = {
 	1,
@@ -128,7 +138,7 @@ store_common_region (cob_field *f, const unsigned char *data,
 	gcf = cob_min_int (hf1, hf2);
 	memset (COB_FIELD_DATA (f), '0', COB_FIELD_SIZE (f));
 	if (gcf > lcf) {
-		csize = (size_t)(gcf - lcf);
+		csize = (size_t)gcf - lcf;
 		p = data + hf1 - gcf;
 		q = COB_FIELD_DATA (f) + hf2 - gcf;
 		for (cinc = 0; cinc < csize; ++cinc, ++p, ++q) {
@@ -153,7 +163,7 @@ cob_binary_mget_sint64 (const cob_field * const f)
 			own_byte_memcpy ((unsigned char *)&n, f->data, f->size);
 			n = COB_BSWAP_64 (n);
 			/* Shift with sign */
-			n >>= 8 * fsiz;
+			n >>= (cob_s64_t)8 * fsiz;
 		} else {
 			own_byte_memcpy (((unsigned char *)&n) + fsiz, f->data, f->size);
 			n = COB_BSWAP_64 (n);
@@ -162,7 +172,7 @@ cob_binary_mget_sint64 (const cob_field * const f)
 		if (COB_FIELD_HAVE_SIGN (f)) {
 			own_byte_memcpy (((unsigned char *)&n) + fsiz, f->data, f->size);
 			/* Shift with sign */
-			n >>= 8 * fsiz;
+			n >>= (cob_s64_t)8 * fsiz;
 		} else {
 			own_byte_memcpy ((unsigned char *)&n, f->data, f->size);
 		}
@@ -495,12 +505,12 @@ cob_move_packed_to_display (cob_field *f1, cob_field *f2)
 
 	/* Unpack string */
 	data = f1->data;
+	offset = COB_FIELD_DIGITS(f1) % 2;
 	if (COB_FIELD_NO_SIGN_NIBBLE (f1)) {
 		sign = 0;
-		offset = COB_FIELD_DIGITS(f1) % 2;
 	} else {
 		sign = cob_packed_get_sign (f1);
-		offset = 1 - (COB_FIELD_DIGITS(f1) % 2);
+		offset = 1 - offset;
 	}
 	for (i = offset; i < COB_FIELD_DIGITS(f1) + offset; ++i) {
 		if (i % 2 == 0) {
@@ -546,27 +556,33 @@ static void
 cob_move_binary_to_binary (cob_field *f1, cob_field *f2)
 {
 	union {
-		cob_u64_t		val;
-		cob_s64_t		val2;
+		cob_u64_t		uval;
+		cob_s64_t		sval;
 	}		ul64;
 	unsigned int	sign;
 
 	sign = 0;
 	if (COB_FIELD_HAVE_SIGN (f1)) {
-		ul64.val2 = cob_binary_mget_sint64 (f1);
-		if (ul64.val2 < 0) {
+		ul64.sval = cob_binary_mget_sint64 (f1);
+		if (ul64.sval < 0) {
 			sign = 1;
 		}
+		if (COB_FIELD_BINARY_TRUNC (f2)) {
+			ul64.sval %= cob_exp10_ll[(int)COB_FIELD_DIGITS(f2)];
+		}
 	} else {
-		ul64.val = cob_binary_mget_uint64 (f1);
+		ul64.uval = cob_binary_mget_uint64 (f1);
+		if (COB_FIELD_BINARY_TRUNC (f2)) {
+			ul64.uval %= cob_exp10_ll[(int)COB_FIELD_DIGITS(f2)];
+		}
 	}
 	if (COB_FIELD_HAVE_SIGN (f2)) {
-		cob_binary_mset_sint64 (f2, ul64.val2);
+		cob_binary_mset_sint64 (f2, ul64.sval);
 	} else {
 		if (sign) {
-			cob_binary_mset_uint64 (f2, (cob_u64_t)(-ul64.val2));
+			cob_binary_mset_uint64 (f2, (cob_u64_t)(-ul64.sval));
 		} else {
-			cob_binary_mset_uint64 (f2, ul64.val);
+			cob_binary_mset_uint64 (f2, ul64.uval);
 		}
 	}
 }
@@ -576,7 +592,6 @@ cob_move_display_to_binary (cob_field *f1, cob_field *f2)
 {
 	unsigned char	*data1;
 	cob_u64_t	val;
-	cob_s64_t	val2;
 	size_t		i, size;
 	size_t		size1;
 	int		sign;
@@ -596,16 +611,16 @@ cob_move_display_to_binary (cob_field *f1, cob_field *f2)
 		}
 	}
 
-	if (COB_FIELD_BINARY_TRUNC (f2) &&
-	    !COB_FIELD_REAL_BINARY(f2)) {
+	if (COB_FIELD_BINARY_TRUNC (f2)) {
 		val %= cob_exp10_ll[(int)COB_FIELD_DIGITS(f2)];
 	}
 
 	if (COB_FIELD_HAVE_SIGN (f2)) {
-		/* Could this cast cause overflows? */
-		val2 = (cob_s64_t)val;
+		cob_s64_t	val2;
 		if (sign < 0) {
-			val2 *= -1;
+			val2 = (cob_s64_t)-1 * val;
+		} else {
+			val2 = (cob_s64_t)1 * val;
 		}
 		cob_binary_mset_sint64 (f2, val2);
 	} else {
@@ -619,7 +634,6 @@ static void
 cob_move_binary_to_display (cob_field *f1, cob_field *f2)
 {
 	cob_u64_t		val;
-	cob_s64_t		val2;
 	int			i;
 	int			sign;
 	char			buff[32];
@@ -627,12 +641,12 @@ cob_move_binary_to_display (cob_field *f1, cob_field *f2)
 	sign = 1;
 	/* Get value */
 	if (COB_FIELD_HAVE_SIGN (f1)) {
-		val2 = cob_binary_mget_sint64 (f1);
+		cob_s64_t		val2 = cob_binary_mget_sint64 (f1);
 		if (val2 < 0) {
 			sign = -1;
-			val = (cob_u64_t)-val2;
+			val = (cob_u64_t)-1 * val2;
 		} else {
-			val = (cob_u64_t)val2;
+			val = (cob_u64_t)1 * val2;
 		}
 	} else {
 		val = cob_binary_mget_uint64 (f1);
@@ -646,7 +660,7 @@ cob_move_binary_to_display (cob_field *f1, cob_field *f2)
 	}
 
 	/* Store */
-	store_common_region (f2, (cob_u8_ptr)buff + i, (size_t)(20 - i),
+	store_common_region (f2, (cob_u8_ptr)buff + i, (size_t)20 - i,
 		COB_FIELD_SCALE(f1));
 
 	COB_PUT_SIGN (f2, sign);
@@ -695,6 +709,10 @@ cob_move_display_to_edited (cob_field *f1, cob_field *f2)
 	}
 
 	/* Count the number of digit places before decimal point */
+	/*
+	  TO-DO: This is computed in cb_build_picture; add computed results to
+	  cb_field and use those.
+	*/
 	for (p = COB_FIELD_PIC (f2); p->symbol; ++p) {
 		c = p->symbol;
 		repeat = p->times_repeated;
@@ -924,18 +942,16 @@ cob_move_display_to_edited (cob_field *f1, cob_field *f2)
 					}
 				}
 				if (sign_symbol && curr_symbol) {
+					/*
+					  Only one of $ and +/- can be floating
+					  in any given picture, so the symbol
+					  which comes after the other must be
+					  the one which floats.
+					*/
 					if (sign_first) {
 						*dst = curr_symbol;
-						--dst;
-						if (dst >= f2->data) {
-							*dst = sign_symbol;
-						}
 					} else {
 						*dst = sign_symbol;
-						--dst;
-						if (dst >= f2->data) {
-							*dst = curr_symbol;
-						}
 					}
 				} else if (sign_symbol) {
 					*dst = sign_symbol;
@@ -1114,15 +1130,14 @@ static void
 cob_move_all (cob_field *src, cob_field *dst)
 {
 	unsigned char		*p;
-	size_t			i;
 	size_t			digcount;
 	cob_field		temp;
-	cob_field_attr		attr;
 
 	if (likely(COB_FIELD_IS_ALNUM(dst))) {
 		if (likely(src->size == 1)) {
 			memset (dst->data, src->data[0], dst->size);
 		} else {
+			size_t			i;
 			digcount = src->size;
 			for (i = 0; i < dst->size; ++i) {
 				dst->data[i] = src->data[i % digcount];
@@ -1130,21 +1145,24 @@ cob_move_all (cob_field *src, cob_field *dst)
 		}
 		return;
 	}
-	COB_ATTR_INIT (COB_TYPE_ALPHANUMERIC, 0, 0, 0, NULL);
-	if (COB_FIELD_IS_NUMERIC(dst)) {
-		digcount = COB_MAX_DIGITS;
-		attr.type = COB_TYPE_NUMERIC_DISPLAY;
-		attr.digits = COB_MAX_DIGITS;
-	} else {
+	if (!COB_FIELD_IS_NUMERIC(dst)) {
+		temp.attr = &all_display_attr;
 		digcount = dst->size;
+	} else if (likely(src->size == 1)) {
+		memset (all_numeric_data, src->data[0], COB_MAX_DIGITS);
+		cob_move ((cob_field *)&all_numeric_field, dst);
+		return;
+	} else {
+		temp.attr = &all_numeric_display_attr;
+		digcount = COB_MAX_DIGITS;	/* CHECKME: when do we enter here? */
 	}
 	p = cob_malloc (digcount);
 	temp.size = digcount;
 	temp.data = p;
-	temp.attr = &attr;
 	if (likely(src->size == 1)) {
 		memset (p, src->data[0], digcount);
 	} else {
+		size_t			i;
 		for (i = 0; i < digcount; ++i) {
 			p[i] = src->data[i % src->size];
 		}
@@ -1163,9 +1181,35 @@ cob_move_ibm (void *dst, void *src, const int len)
 {
 	char	*dest = dst;
 	char	*srce = src;
-	int	i;
-	for(i=0; i < len; i++)
-		dest[i] = srce[i];
+	int		i = len;
+	while(i-- > 0) {
+		*dest = *srce;
+		dest++;
+		srce++;
+	}
+}
+
+/*
+ * Propagate table (1) throughout the table
+ * (used by INITIALIZE)
+ */
+void
+cob_init_table (void *tbl, const size_t len, const size_t occ)
+{
+	char	*m = (char*)tbl + len;
+	size_t	i = len;
+	size_t	j = 1;
+	if (occ < 1)
+		return;
+	do {
+		j = j * 2;
+		memcpy((void*)m, tbl, i);
+		m = m + i;
+		i = i * 2;
+	} while ((j * 2) < occ);
+	if (j < occ) {
+		memcpy((void*)m, tbl, len * (occ - j));
+	}
 }
 
 void
@@ -1195,16 +1239,16 @@ cob_move (cob_field *src, cob_field *dst)
 	}
 
 	/* Non-elementary move */
-	if (COB_FIELD_TYPE (src) == COB_TYPE_GROUP ||
-	    COB_FIELD_TYPE (dst) == COB_TYPE_GROUP) {
+	if (COB_FIELD_TYPE (src) == COB_TYPE_GROUP
+	 || COB_FIELD_TYPE (dst) == COB_TYPE_GROUP) {
 		cob_move_alphanum_to_alphanum (src, dst);
 		return;
 	}
 
 	opt = 0;
 	if (COB_FIELD_TYPE (dst) == COB_TYPE_NUMERIC_BINARY) {
-		if (COB_FIELD_BINARY_TRUNC (dst) &&
-		    !COB_FIELD_REAL_BINARY(dst)) {
+		if (COB_FIELD_BINARY_TRUNC (dst)
+		 && !COB_FIELD_REAL_BINARY (dst)) {
 			opt = COB_STORE_TRUNC_ON_OVERFLOW;
 		}
 	}
@@ -1230,18 +1274,19 @@ cob_move (cob_field *src, cob_field *dst)
 			cob_move_display_to_packed (src, dst);
 			return;
 		case COB_TYPE_NUMERIC_BINARY:
+		case COB_TYPE_NUMERIC_COMP5:
 			cob_move_display_to_binary (src, dst);
 			return;
 		case COB_TYPE_NUMERIC_EDITED:
 			cob_move_display_to_edited (src, dst);
 			return;
 		case COB_TYPE_ALPHANUMERIC_EDITED:
-			if (COB_FIELD_SCALE(src) < 0 ||
-			    COB_FIELD_SCALE(src) > COB_FIELD_DIGITS(src)) {
+			if (COB_FIELD_SCALE (src) < 0
+			 || COB_FIELD_SCALE (src) > COB_FIELD_DIGITS (src)) {
 				/* Expand P's */
 				indirect_move (cob_move_display_to_display, src, dst,
-					      (size_t)cob_max_int ((int)COB_FIELD_DIGITS(src), (int)COB_FIELD_SCALE(src)),
-					      cob_max_int (0, (int)COB_FIELD_SCALE(src)));
+						(size_t)cob_max_int ((int)COB_FIELD_DIGITS(src), (int)COB_FIELD_SCALE(src)),
+						cob_max_int (0, (int)COB_FIELD_SCALE(src)));
 				return;
 			} else {
 				cob_move_alphanum_to_edited (src, dst);
@@ -1258,6 +1303,7 @@ cob_move (cob_field *src, cob_field *dst)
 			cob_move_packed_to_display (src, dst);
 			return;
 		case COB_TYPE_NUMERIC_BINARY:
+		case COB_TYPE_NUMERIC_COMP5:
 			cob_decimal_setget_fld (src, dst, opt);
 			return;
 		case COB_TYPE_NUMERIC_PACKED:
@@ -1279,8 +1325,10 @@ cob_move (cob_field *src, cob_field *dst)
 		}
 
 	case COB_TYPE_NUMERIC_BINARY:
+	case COB_TYPE_NUMERIC_COMP5:
 		switch (COB_FIELD_TYPE (dst)) {
 		case COB_TYPE_NUMERIC_BINARY:
+		case COB_TYPE_NUMERIC_COMP5:
 			if (COB_FIELD_SCALE(src) == COB_FIELD_SCALE(dst)) {
 				cob_move_binary_to_binary (src, dst);
 				return;
@@ -1303,13 +1351,13 @@ cob_move (cob_field *src, cob_field *dst)
 			return;
 		case COB_TYPE_NUMERIC_EDITED:
 			indirect_move (cob_move_binary_to_display, src, dst,
-				       (size_t)COB_MAX_DIGITS,
-				       COB_FIELD_SCALE(src));
+					(size_t)COB_MAX_DIGITS,
+					COB_FIELD_SCALE(src));
 			return;
 		default:
 			indirect_move (cob_move_binary_to_display, src, dst,
-				       (size_t)(COB_FIELD_DIGITS(src)),
-				       COB_FIELD_SCALE(src));
+					(size_t)(COB_FIELD_DIGITS(src)),
+					COB_FIELD_SCALE(src));
 			return;
 		}
 
@@ -1320,6 +1368,7 @@ cob_move (cob_field *src, cob_field *dst)
 			return;
 		case COB_TYPE_NUMERIC_PACKED:
 		case COB_TYPE_NUMERIC_BINARY:
+		case COB_TYPE_NUMERIC_COMP5:
 		case COB_TYPE_NUMERIC_EDITED:
 		case COB_TYPE_NUMERIC_FLOAT:
 		case COB_TYPE_NUMERIC_DOUBLE:
@@ -1350,6 +1399,7 @@ cob_move (cob_field *src, cob_field *dst)
 			cob_move_fp_to_fp (src, dst);
 			return;
 		case COB_TYPE_NUMERIC_BINARY:
+		case COB_TYPE_NUMERIC_COMP5:
 			cob_decimal_setget_fld (src, dst, opt);
 			return;
 		case COB_TYPE_NUMERIC_PACKED:
@@ -1376,6 +1426,7 @@ cob_move (cob_field *src, cob_field *dst)
 			cob_move_fp_to_fp (src, dst);
 			return;
 		case COB_TYPE_NUMERIC_BINARY:
+		case COB_TYPE_NUMERIC_COMP5:
 			cob_decimal_setget_fld (src, dst, opt);
 			return;
 		case COB_TYPE_NUMERIC_PACKED:
@@ -1396,6 +1447,7 @@ cob_move (cob_field *src, cob_field *dst)
 	case COB_TYPE_NUMERIC_FP_DEC64:
 		switch (COB_FIELD_TYPE (dst)) {
 		case COB_TYPE_NUMERIC_BINARY:
+		case COB_TYPE_NUMERIC_COMP5:
 			cob_decimal_setget_fld (src, dst, opt);
 			return;
 		case COB_TYPE_NUMERIC_FP_DEC64:
@@ -1418,6 +1470,7 @@ cob_move (cob_field *src, cob_field *dst)
 	case COB_TYPE_NUMERIC_FP_DEC128:
 		switch (COB_FIELD_TYPE (dst)) {
 		case COB_TYPE_NUMERIC_BINARY:
+		case COB_TYPE_NUMERIC_COMP5:
 			cob_decimal_setget_fld (src, dst, opt);
 			return;
 		case COB_TYPE_NUMERIC_FP_DEC128:
@@ -1445,6 +1498,7 @@ cob_move (cob_field *src, cob_field *dst)
 			return;
 		case COB_TYPE_NUMERIC_PACKED:
 		case COB_TYPE_NUMERIC_BINARY:
+		case COB_TYPE_NUMERIC_COMP5:
 		case COB_TYPE_NUMERIC_EDITED:
 		case COB_TYPE_NUMERIC_FLOAT:
 		case COB_TYPE_NUMERIC_DOUBLE:
@@ -1455,7 +1509,7 @@ cob_move (cob_field *src, cob_field *dst)
 		case COB_TYPE_NUMERIC_FP_DEC64:
 		case COB_TYPE_NUMERIC_FP_DEC128:
 			indirect_move (cob_move_alphanum_to_display, src, dst,
-					(size_t)(2* COB_MAX_DIGITS),
+					(size_t)(2 * COB_MAX_DIGITS),
 					COB_MAX_DIGITS);
 			return;
 		case COB_TYPE_ALPHANUMERIC_EDITED:
@@ -1480,12 +1534,12 @@ cob_packed_get_int (cob_field *f1)
 	int		sign;
 
 	data = f1->data;
+	offset = COB_FIELD_DIGITS(f1) % 2;
 	if (COB_FIELD_NO_SIGN_NIBBLE (f1)) {
 		sign = 0;
-		offset = COB_FIELD_DIGITS(f1) % 2;
 	} else {
 		sign = cob_packed_get_sign (f1);
-		offset = 1 - (COB_FIELD_DIGITS(f1) % 2);
+		offset = 1 - offset;
 	}
 	for (i = offset; i < COB_FIELD_DIGITS(f1) - COB_FIELD_SCALE(f1) + offset; ++i) {
 		val *= 10;
@@ -1507,18 +1561,20 @@ cob_packed_get_long_long (cob_field *f1)
 	unsigned char	*data;
 	size_t		i;
 	size_t		offset;
+	size_t		field_data;
 	cob_s64_t	val = 0;
 	int		sign;
 
 	data = f1->data;
+	offset = COB_FIELD_DIGITS(f1) % 2;
 	if (COB_FIELD_NO_SIGN_NIBBLE (f1)) {
 		sign = 0;
-		offset = COB_FIELD_DIGITS(f1) % 2;
 	} else {
 		sign = cob_packed_get_sign (f1);
-		offset = 1 - (COB_FIELD_DIGITS(f1) % 2);
+		offset = 1 - offset;
 	}
-	for (i = offset; i < COB_FIELD_DIGITS(f1) - COB_FIELD_SCALE(f1) + offset; ++i) {
+	field_data = COB_FIELD_DIGITS(f1) - COB_FIELD_SCALE(f1);
+	for (i = offset; i < field_data + offset; ++i) {
 		val *= 10;
 		if (i % 2 == 0) {
 			val += data[i / 2] >> 4;
@@ -1638,6 +1694,7 @@ cob_get_int (cob_field *f)
 	case COB_TYPE_NUMERIC_PACKED:
 		return cob_packed_get_int (f);
 	case COB_TYPE_NUMERIC_BINARY:
+	case COB_TYPE_NUMERIC_COMP5:
 		val = cob_binary_mget_sint64 (f);
 		for (n = COB_FIELD_SCALE (f); n > 0 && val; --n) {
 			val /= 10;
@@ -1667,6 +1724,7 @@ cob_get_llint (cob_field *f)
 	case COB_TYPE_NUMERIC_PACKED:
 		return cob_packed_get_long_long (f);
 	case COB_TYPE_NUMERIC_BINARY:
+	case COB_TYPE_NUMERIC_COMP5:
 		n = cob_binary_mget_sint64 (f);
 		for (inc = COB_FIELD_SCALE (f); inc > 0 && n; --inc) {
 			n /= 10;
@@ -1694,7 +1752,9 @@ cob_init_move (cob_global *lptr, cob_settings *sptr)
 void
 cob_put_u64_compx (cob_u64_t val, void *mem, int len)
 {
+#if !defined(WORDS_BIGENDIAN)
 	cob_u64_t	ulong;
+#endif
 	cob_u32_t	uint;
 	cob_u16_t	ushort;
 	
@@ -1786,7 +1846,9 @@ cob_put_u64_comp5 (cob_u64_t val, void *mem, int len)
 void
 cob_put_s64_compx (cob_s64_t val, void *mem, int len)
 {
+#if !defined(WORDS_BIGENDIAN)
 	cob_s64_t	slong;
+#endif
 	cob_s32_t	sint;
 	cob_s16_t	sshort;
 #if defined(WORDS_BIGENDIAN)
@@ -2340,13 +2402,12 @@ cob_put_pointer (void *val, void *mem)
 }
 
 char *
-cob_get_picx (void *cbl_data, int len, void *char_field, int num_chars)
+cob_get_picx (void *cbl_data, size_t len, void *char_field, size_t num_chars)
 {
-	int		i;
+	size_t		i;
 	cob_u8_t	*p = cbl_data;
 
-	for (i = len - 1; i >= 0 && (p[i] == ' ' || p[i] == 0); i--);
-	i++;
+	for (i = len; i != 0 && (p[i - 1] == ' ' || p[i - 1] == 0); i--);
 
 	if (char_field == NULL) {
 		num_chars = i + 1;
@@ -2363,9 +2424,9 @@ cob_get_picx (void *cbl_data, int len, void *char_field, int num_chars)
 }
 
 void
-cob_put_picx( void *cbl_data, int len, void *string)
+cob_put_picx( void *cbl_data, size_t len, void *string)
 {
-	int	i,j;
+	size_t	i, j;
 	cob_u8_t	*p = cbl_data;
 	j = strlen ((char*)string);
 	if (j > len) {
